package net.wanji.utc.task;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import lombok.extern.slf4j.Slf4j;
import net.wanji.common.framework.rest.JsonViewObject;
import net.wanji.common.utils.tool.JacksonUtils;
import net.wanji.databus.dao.entity.CrossLightsStatusHistPO;
import net.wanji.databus.dao.entity.SignalCommandLogPO;
import net.wanji.databus.dao.entity.SignalCommandPO;
import net.wanji.databus.dao.mapper.CrossInfoMapper;
import net.wanji.databus.dao.mapper.CrossLightsStatusHistMapper;
import net.wanji.databus.dao.mapper.ManufacturerInfoMapper;
import net.wanji.databus.po.CrossInfoPO;
import net.wanji.databus.po.ManufacturerInfoPO;
import net.wanji.databus.vo.LightsStatusVO;
import net.wanji.databus.vo.LightsStatusVO2;
import net.wanji.feign.service.EHualuFeignClients;
import net.wanji.feign.service.UtcDTFeignClients;
import net.wanji.feign.service.UtcHisenseFeignClients;
import net.wanji.utc.common.constant.Constants;
import net.wanji.utc.common.typeenum.BasicEnum;
import net.wanji.utc.kafka.ProducerHandler;
import net.wanji.utc.service.runninginfo.HkLightsStatusService;
import net.wanji.utc.service.runninginfo.SignalStatusService;
import net.wanji.utc.websocket.RealTimeDataWebSocket;
import org.apache.commons.lang3.StringUtils;
import org.elasticsearch.action.index.IndexRequest;
import org.elasticsearch.action.index.IndexResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.client.indices.CreateIndexRequest;
import org.elasticsearch.client.indices.CreateIndexResponse;
import org.elasticsearch.client.indices.GetIndexRequest;
import org.elasticsearch.common.xcontent.XContentType;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.scheduling.concurrent.ThreadPoolTaskExecutor;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

import javax.annotation.Resource;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;

/**
 * 订阅信号运行状态、告警、灯态
 *
 * @author Kent HAN
 * @date 2022/11/21 9:01
 */
@Component
@Slf4j
@SuppressWarnings("all")
public class SignalStatusTask {
    @Autowired
    SignalStatusService signalStatusService;
    @Resource(name = "commonThreadPoolExecutor")
    ThreadPoolTaskExecutor commonThreadPoolExecutor;

    @Resource
    private HkLightsStatusService hkLightsStatusService;
    @Resource
    private ManufacturerInfoMapper manufacturerInfoMapper;
    @Resource
    private CrossInfoMapper crossInfoMapper;
    @Resource
    private UtcDTFeignClients utcDTFeignClients;
    @Resource
    private CrossLightsStatusHistMapper crossLightsStatusHistMapper;
    @Autowired
    private ProducerHandler producerHandler;
    @Resource
    RestHighLevelClient client;

    private static final Map<String, Integer> runModeTimeMap = new ConcurrentHashMap<>();
    @Qualifier("net.wanji.feign.service.UtcHisenseFeignClients")
    @Autowired
    private UtcHisenseFeignClients utcHisenseFeignClients;
    @Qualifier("net.wanji.feign.service.EHualuFeignClients")
    @Autowired
    private EHualuFeignClients eHualuFeignClients;

    public static final ConcurrentHashMap<String, List<LightsStatusVO2>> produceListMap = new ConcurrentHashMap<>();

    public static final ConcurrentHashMap<String, SignalCommandPO> crossLightsStatusMap = new ConcurrentHashMap<>();

    // 运行状态、告警，1分钟一次
    @Scheduled(fixedRate = 60 * 1000)
    public void runningStatusAlarm() throws Exception {
        signalStatusService.runningStatusAlarm();
    }

    // 灯态，每秒一次
    @Scheduled(fixedRate = 1000)
    public void lightStatus() {
        ConcurrentHashMap<String, RealTimeDataWebSocket> evenWarnWebSocketMap = RealTimeDataWebSocket.getEvenWarnWebSocketMap();
        try {
            if (!evenWarnWebSocketMap.isEmpty()) {
                Map<String, List<Map.Entry<String, RealTimeDataWebSocket>>> crossEntryMap = getCrossEntryMap(evenWarnWebSocketMap);
                for (Map.Entry<String, List<Map.Entry<String, RealTimeDataWebSocket>>> entry : crossEntryMap.entrySet()) {
                    commonThreadPoolExecutor.execute(() -> {
                        long start = System.currentTimeMillis();
                        ObjectMapper mapper = JacksonUtils.getInstance();
                        String crossId = entry.getKey();
                        CrossInfoPO crossInfoPO = crossInfoMapper.selectByPrimaryKey(crossId);
                        if (Objects.nonNull(crossInfoPO)) {
                            List<LightsStatusVO> lightsStatusVOS = getLightsStatusVOS(mapper, crossId, crossInfoPO);
                            setStatus(crossId, start, lightsStatusVOS);
                            saveAndSendLightsStatus(mapper, crossId, lightsStatusVOS, entry);
                            long end = System.currentTimeMillis();
                            log.error("当前路口：{}, 获取灯态时间：{}ms", crossId, end-start);
                        }
                    });
                }
                sendKafka();
                //sendElastic();
            }
        } catch (Exception e) {
            log.error("实时推送灯态数据失败", e);
        }
    }

    private static void setStatus(String crossId, long start, List<LightsStatusVO> lightsStatusVOS) {
        // 添加在线离线逻辑
        // 0-离线 1-在线
        SignalCommandLogPO signalCommandLogPO = new SignalCommandLogPO();
        int status = 0;
        String runMode = "11";
        String machineTimestamp = null;
        if (!CollectionUtils.isEmpty(lightsStatusVOS)) {
            for (LightsStatusVO lightsStatusVO : lightsStatusVOS) {
                if (Objects.nonNull(lightsStatusVO)) {
                    machineTimestamp = lightsStatusVO.getTimeStamp();
                    long offsetTime = start - Long.valueOf(machineTimestamp);
                    if (offsetTime < 1000 * 12) {
                        status = 1;
                    }
                    runMode = lightsStatusVO.getRunMode();
                    log.error("当前路口：{}，当前信号机状态时间差：{}， 当前控制模式:{}, 当前在线离线状态：{}",
                            crossId, offsetTime, runMode, status);
                }
            }
        }
        signalCommandLogPO.setCrossId(crossId);
        signalCommandLogPO.setRunMode(runMode);
        signalCommandLogPO.setStatus(status);
        signalCommandLogPO.setUpdateTime(new Date());
        signalCommandLogPO.setLampTime(machineTimestamp);
        crossLightsStatusMap.put(crossId, signalCommandLogPO);
    }

    /**
     * 保存灯态数据并发送
     *
     * @param mapper
     * @param crossId
     * @param lightsStatusVOS
     */
    private void saveAndSendLightsStatus(ObjectMapper mapper, String crossId, List<LightsStatusVO> lightsStatusVOS,
                                         Map.Entry<String, List<Map.Entry<String, RealTimeDataWebSocket>>> entry) {
        List<Map.Entry<String, RealTimeDataWebSocket>> value = entry.getValue();
        produceListMap.remove(crossId); // 清除上一秒缓存
        if (!CollectionUtils.isEmpty(lightsStatusVOS)) {
            for (LightsStatusVO lightsStatusVO : lightsStatusVOS) {
                if (Objects.nonNull(lightsStatusVO) && Objects.equals(lightsStatusVO.getCrossId(), crossId)) {
                    calculateControlCountDown(crossId, lightsStatusVO);
                    try {
                        // 相同路口不同websocket统一发送灯态
                        List<LightsStatusVO2> listResult = convertDirLampGroupMap2Entity(lightsStatusVOS);
                        String json = mapper.writeValueAsString(listResult);
                        for (Map.Entry<String, RealTimeDataWebSocket> socketEntry : value) {
                            String crossIdStr = socketEntry.getKey();
                            RealTimeDataWebSocket webSocket = socketEntry.getValue();
                            webSocket.sendInfo(json, crossIdStr);
                        }
                        insertIntoHist(crossId, json);
                        produceListMap.put(crossId, listResult);
                    } catch (Exception e) {
                        log.error("RealTimeDataWebSocket发送异常，异常信息：", e);
                    }
                }
            }
        }
    }

    /**
     * 获取不同厂商的灯态列表
     *
     * @param mapper
     * @param crossId
     * @param manufacturerInfoPO
     * @return
     */
    private List<LightsStatusVO> getLightsStatusVOS(ObjectMapper mapper, String crossId, CrossInfoPO crossInfoPO) {
        try {
            ManufacturerInfoPO manufacturerInfoPO = manufacturerInfoMapper.selectById(crossInfoPO.getManufacturerId());
            if (StringUtils.equals(BasicEnum.ManufacturerEnum.HK.getCode(), manufacturerInfoPO.getCode())) {
                return hkLightsStatusService.getHkLightsStatus(crossId);
            }
            if (StringUtils.equals(BasicEnum.ManufacturerEnum.DT.getCode(), manufacturerInfoPO.getCode())) {
                JsonViewObject jsonViewObject = utcDTFeignClients.lightStatus(crossId);
                return mapper.convertValue(jsonViewObject.getContent(), new TypeReference<List<LightsStatusVO>>() {
                });
            }
            if (StringUtils.endsWithIgnoreCase(BasicEnum.ManufacturerEnum.HISENSE.getCode(), manufacturerInfoPO.getCode())) {
                JsonViewObject jsonViewObject = utcHisenseFeignClients.lightStatus(crossId);
                return mapper.convertValue(jsonViewObject.getContent(), new TypeReference<List<LightsStatusVO>>() {
                });
            }
            if (StringUtils.endsWithIgnoreCase(BasicEnum.ManufacturerEnum.EHULU.getCode(), manufacturerInfoPO.getCode())) {
                JsonViewObject jsonViewObject = eHualuFeignClients.lightStatus(crossId);
                return mapper.convertValue(jsonViewObject.getContent(), new TypeReference<List<LightsStatusVO>>() {
                });
            }
            if (StringUtils.endsWithIgnoreCase("STATIC", manufacturerInfoPO.getCode())) {
                LightsStatusVO lightsStatusVO = SignalStatus4StaticSchemeTask.currentCrossLightStatusCache.get(crossId);
                List<LightsStatusVO> lightsStatusVOS = new ArrayList<>();
                if (Objects.nonNull(lightsStatusVO)) {
                    lightsStatusVOS.add(lightsStatusVO);
                }
                return lightsStatusVOS;
            }
        } catch (Exception e) {
            log.error("获取当前厂商路口编号[{}]灯态异常", crossId, e);
        }
        return Collections.emptyList();
    }

    private static List<LightsStatusVO2> convertDirLampGroupMap2Entity(List<LightsStatusVO> lightsStatusVOS) throws Exception {
        List<LightsStatusVO2> listResult = null;
        try {
            listResult = new ArrayList<>(lightsStatusVOS.size());
            for (LightsStatusVO lightsStatusVO : lightsStatusVOS) {
                LightsStatusVO2 lightsStatusVO2 = new LightsStatusVO2();
                BeanUtils.copyProperties(lightsStatusVO, lightsStatusVO2);
                Map<String, Object> dirLampGroupMap = lightsStatusVO.getDirLampGroupMap();
                List<LightsStatusVO2.DirInfo> dirList = new ArrayList<>();
                Map<String, Object> phaseMap = lightsStatusVO.getPhaseMap();
                if (Objects.isNull(phaseMap) || phaseMap.isEmpty()) {
                    continue;
                }
                for (Map.Entry<String, Object> groupEntry : dirLampGroupMap.entrySet()) {
                    LightsStatusVO2.DirInfo dirInfo = new LightsStatusVO2.DirInfo();
                    String dir = groupEntry.getKey();
                    Map<Integer, String> turnColorMap = (Map<Integer, String>) groupEntry.getValue();
                    List<LightsStatusVO2.TurnInfo> turnList = new ArrayList<>();
                    for (Map.Entry<Integer, String> turnEntry : turnColorMap.entrySet()) {
                        setDirTurnInfo(phaseMap, dir, turnList, turnEntry);
                    }
                    dirInfo.setDir(dir);
                    dirInfo.setTurnList(turnList);
                    dirList.add(dirInfo);
                }
                lightsStatusVO2.setDirLampGroupMapList(dirList);
                listResult.add(lightsStatusVO2);
            }
        } catch (Exception e) {
            log.error("灯态推送转换V2异常：", e);
            throw new Exception(e);
        }
        return listResult;
    }

    /**
     * 设置方向转向信息实体
     *
     * @param phaseMap
     * @param dir
     * @param turnList
     * @param turnEntry
     */
    private static void setDirTurnInfo(Map<String, Object> phaseMap, String dir, List<LightsStatusVO2.TurnInfo> turnList, Map.Entry<Integer, String> turnEntry) {
        Integer turn = parseTurn(turnEntry);
        String color = turnEntry.getValue();
        LightsStatusVO2.TurnInfo turnInfo = new LightsStatusVO2.TurnInfo();
        turnInfo.setTurn(String.valueOf(turn));
        if (Objects.equals(turn, 20) && Objects.equals(color, "yellow")) { //行人黄灯时。绿闪
            turnInfo.setColor("greenFlash");
        } else {
            turnInfo.setColor(color);
        }
        setTurnInfoCounDown(phaseMap, dir, turn, turnInfo);
        turnList.add(turnInfo);
    }

    /**
     * 转换转向参数
     *
     * @param turnEntry
     * @return
     */
    private static Integer parseTurn(Map.Entry<Integer, String> turnEntry) {
        Integer turn = null;
        if (turnEntry.getKey() instanceof Integer) {
            turn = (Integer) turnEntry.getKey();
        } else {
            Object object = turnEntry.getKey();
            turn = Integer.parseInt(String.valueOf(object));
        }
        return turn;
    }

    /**
     * 设置方向转向倒计时时间
     *
     * @param phaseMap
     * @param dir
     * @param turn
     * @param turnInfo
     */
    private static void setTurnInfoCounDown(Map<String, Object> phaseMap, String dir, Integer turn, LightsStatusVO2.TurnInfo turnInfo) {
        String key = dir.concat(Constants.SEPARATOR_UNDER_LINE).concat(String.valueOf(turn));
        if (phaseMap.get(key) instanceof Map) {
            Map<Integer, Integer> turnMap = (Map<Integer, Integer>) phaseMap.get(key);
            if (Objects.nonNull(turnMap) && !turnMap.isEmpty()) {
                Integer countDown = turnMap.get(turn);
                turnInfo.setCountDown(countDown);
            }
        } else {
            Integer countDown = null;
            if (phaseMap.get(dir) instanceof Map) {
                Map<String, Integer> turnMap = (Map<String, Integer>) phaseMap.get(dir);
                countDown = turnMap.get(String.valueOf(turn));
            } else {
                countDown = (Integer) phaseMap.get(key);
            }
            turnInfo.setCountDown(countDown);
        }
    }

    /**
     * 灯态数据发送kafka
     */
    private void sendKafka() {
        if (!produceListMap.isEmpty()) {
            List<LightsStatusVO2> list = new ArrayList<>(produceListMap.size());
            for (Map.Entry<String, List<LightsStatusVO2>> entry : produceListMap.entrySet()) {
                list.addAll(entry.getValue());
            }
            producerHandler.send(Constants.SIGNAL_LIGHT_HISTORY_KEY, list);
        }
    }

    /**
     * 灯态数据发送elastic
     */
    private void sendElastic() throws Exception {
        if (!produceListMap.isEmpty()) {
            List<LightsStatusVO2> list = new ArrayList<>(produceListMap.size());
            for (Map.Entry<String, List<LightsStatusVO2>> entry : produceListMap.entrySet()) {
                List<LightsStatusVO2> value = entry.getValue();
                for (LightsStatusVO2 lightsStatusVO2 : value) {
                    String json = JacksonUtils.getInstance().writeValueAsString(lightsStatusVO2);
                    GetIndexRequest getIndexRequest = new GetIndexRequest(Constants.SIGNAL_LIGHT_HISTORY_KEY);
                    boolean exists = client.indices().exists(getIndexRequest, RequestOptions.DEFAULT);
                    if (exists) {
                        IndexRequest indexRequest = new IndexRequest(Constants.SIGNAL_LIGHT_HISTORY_KEY);
                        indexRequest.source(json, XContentType.JSON);
                        IndexResponse response = client.index(indexRequest, RequestOptions.DEFAULT);
                    } else {
                        CreateIndexRequest createIndexRequest = new CreateIndexRequest(Constants.SIGNAL_LIGHT_HISTORY_KEY);
                        createIndexRequest.source(json, XContentType.JSON);
                        CreateIndexResponse createIndexResponse = client.indices().create(createIndexRequest, RequestOptions.DEFAULT);
                    }
                }
            }
        }
    }

    /**
     * 计算控制倒计时
     *
     * @param crossId
     * @param lightsStatusVO
     */
    private static void calculateControlCountDown(String crossId, LightsStatusVO lightsStatusVO) {
        Integer countTime = 0;
        // 特殊模式控制逻辑
        if (Arrays.asList(2, 3, 4, 9).contains(Integer.valueOf(lightsStatusVO.getRunMode()))) {
            if (runModeTimeMap.isEmpty() || Objects.isNull(runModeTimeMap.get(crossId))) {
                runModeTimeMap.put(crossId, countTime);
            } else {
                Integer countTimeCache = runModeTimeMap.get(crossId);
                runModeTimeMap.put(crossId, ++countTimeCache);
            }
            lightsStatusVO.setCycleCountDown(runModeTimeMap.get(crossId));
        } else if (StringUtils.equals(lightsStatusVO.getRunMode(), "11")) {
            runModeTimeMap.remove(crossId);
        }
        Constants.controlModeMap.put(crossId, lightsStatusVO.getRunMode());
    }

    /**
     * 不同系统获取统一路口灯态，统一处理
     *
     * @param evenWarnWebSocketMap
     * @return
     */
    private static Map<String, List<Map.Entry<String, RealTimeDataWebSocket>>> getCrossEntryMap(ConcurrentHashMap<String, RealTimeDataWebSocket> evenWarnWebSocketMap) {
        Map<String, List<Map.Entry<String, RealTimeDataWebSocket>>> tempMap = new ConcurrentHashMap<>();
        for (Map.Entry<String, RealTimeDataWebSocket> item : evenWarnWebSocketMap.entrySet()) {
            String crossId = "";
            String crossIdStr = item.getKey();
            if (StringUtils.isNotBlank(crossIdStr)) {
                String[] split = crossIdStr.split(",");
                crossId = split[0];
            }
            List<Map.Entry<String, RealTimeDataWebSocket>> socketList = new CopyOnWriteArrayList<>();
            if (tempMap.containsKey(crossId)) {
                socketList = tempMap.get(crossId);
                socketList.add(item);
            } else {
                socketList.add(item);
            }
            tempMap.put(crossId, socketList);
        }
        return tempMap;
    }

    /**
     * 插入路口灯态历史表
     *
     * @param crossId 路口ID
     * @param json    灯态JSON字符串
     * @throws Exception 异常处理
     */
    private void insertIntoHist(String crossId, String json) throws Exception {
        try {
            CrossLightsStatusHistPO crossLightsStatusHistPO = new CrossLightsStatusHistPO();
            crossLightsStatusHistPO.setCrossId(crossId);
            LocalDateTime now = LocalDateTime.now();
            Integer batchTime = (int) now.toEpochSecond(ZoneOffset.of("+8"));
            crossLightsStatusHistPO.setBatchTime(batchTime);
            crossLightsStatusHistPO.setLightsStatusJson(json);
            crossLightsStatusHistMapper.insertOne(crossLightsStatusHistPO);
        } catch (Exception e) {
            log.error("插入路口灯态历史表异常", e);
            throw new Exception(e);
        }
    }
}
